/*
Note: this code was written AS QUICKLY AS POSSIBLE.

As such, I barely check return codes.

It was a conscious decision.

This is not production code.

Do not follow this example.
*/
#include <windows.h>
#include <stdio.h>
#include <d3d10.h>
#include <d3dx10.h>

#include <CountPtr.h>

int g_numSamples = 1;

#define countof(a) (sizeof(a)/sizeof((a)[0]))

CountPtr<ID3D10Device> g_device;
CountPtr<IDXGISwapChain> g_swapChain;

CountPtr<ID3D10InputLayout> g_layout;

CountPtr<ID3D10Texture2D> g_backBuffer;
CountPtr<ID3D10RenderTargetView> g_renderTargetView;

CountPtr<ID3D10Buffer> g_constantBuffer;
CountPtr<ID3D10Buffer> g_lightConstantBuffer;

CountPtr<ID3D10Texture2D> g_depthStencilTexture;
CountPtr<ID3D10DepthStencilView> g_depthStencilView;

CountPtr<ID3D10Effect> g_effect;
ID3D10EffectTechnique* g_geometryTechnique, *g_testTechnique, *g_lightTechnique;
DXGI_SWAP_CHAIN_DESC g_swapChainDesc;
D3D10_VIEWPORT g_viewport;

#define NAME  L"D3D10Deferred"
#define TITLE L"D3D10 Deferred Shading Test"

#define LightType_Ambient     0
#define LightType_Directional 1

#define Max_Lights 8

// Light Constant Buffer Information ////////////////////////////////////////////////////////////////////////////////
struct LightData
{
  int type;
  D3DXVECTOR3 positionOrDirection;
  D3DXVECTOR3 color;
  float pad; // Padding to match the size of this struct up to the one in the shader
  
  static LightData Ambient(D3DXVECTOR3 color)
  {
    LightData ld;
    ld.type = LightType_Ambient;
    ld.positionOrDirection = D3DXVECTOR3(0,0,0);
    ld.color = color;
    return ld;
  } 
  
  static LightData Directional(D3DXVECTOR3 direction, D3DXVECTOR3 color)
  {
    LightData ld;
    ld.type = LightType_Directional;
    ld.color = color;
    ld.positionOrDirection = direction;
    return ld;
  }
};

struct LightConstantBuffer
{
  LightData lights[Max_Lights];
  int numLights;
  int padding[3]; // Pad out to a multiple of 16
};

// Texture Helper Class /////////////////////////////////////////////////////////////////////////////////////////////
class Texture : public Counted
{
  CountPtr<ID3D10Texture2D> m_texture;
  CountPtr<ID3D10ShaderResourceView> m_shaderResourceView;
  
    public:
  void Apply(UINT slot)
  {
    g_device->PSSetShaderResources(slot, 1, m_shaderResourceView.ConstAddress());
  }
  
  ID3D10ShaderResourceView *View() { return m_shaderResourceView; }
  
  Texture(const wchar_t *filename)
  {
    CountPtr<ID3D10Resource> resource;
    D3DX10CreateTextureFromFile( g_device, filename, NULL, NULL, resource.Address(), NULL);
    
    resource->QueryInterface(IID_ID3D10Texture2D, (void**)m_texture.Address());
    D3D10_TEXTURE2D_DESC desc;
    m_texture->GetDesc(&desc);
    
    D3D10_SHADER_RESOURCE_VIEW_DESC srDesc;
    srDesc.Format = desc.Format;
    srDesc.ViewDimension = D3D10_SRV_DIMENSION_TEXTURE2D;
    srDesc.Texture2D.MipLevels = desc.MipLevels;
    srDesc.Texture2D.MostDetailedMip = 0;
    g_device->CreateShaderResourceView(m_texture, &srDesc, m_shaderResourceView.Address());
  }
};

// Render Target Texture Helper Class ///////////////////////////////////////////////////////////////////////////////
class RenderTargetTexture : public Counted
{
  CountPtr<ID3D10Texture2D> m_rt;
  CountPtr<ID3D10ShaderResourceView> m_shaderResourceView;
  CountPtr<ID3D10RenderTargetView> m_renderTargetView;
 
    public:
  RenderTargetTexture(int width, int height, DXGI_FORMAT format)
  {
    D3D10_TEXTURE2D_DESC desc;
    desc.Width  = width;
    desc.Height = height;
    desc.MipLevels = 1;
    desc.ArraySize = 1;
    desc.Format = format;
    desc.SampleDesc.Count = g_numSamples;
    desc.SampleDesc.Quality = 0;
    desc.Usage = D3D10_USAGE_DEFAULT;
    desc.BindFlags = D3D10_BIND_RENDER_TARGET | D3D10_BIND_SHADER_RESOURCE;
    desc.CPUAccessFlags = 0;
    desc.MiscFlags = 0;
    
    
    HRESULT hr = g_device->CreateTexture2D(&desc, NULL, m_rt.Address());
    
    if(SUCCEEDED(hr))
    {
      D3D10_RENDER_TARGET_VIEW_DESC rtDesc;
      rtDesc.Format = desc.Format;
      if(desc.SampleDesc.Count != 1)
        rtDesc.ViewDimension = D3D10_RTV_DIMENSION_TEXTURE2DMS;
      else
      {
        rtDesc.ViewDimension = D3D10_RTV_DIMENSION_TEXTURE2D; //MS;
        rtDesc.Texture2D.MipSlice = 0;
      }
      hr = g_device->CreateRenderTargetView( m_rt, &rtDesc, m_renderTargetView.Address());
    }
    
    if(SUCCEEDED(hr))
    {
      D3D10_SHADER_RESOURCE_VIEW_DESC svDesc;
      svDesc.Format = format;
      if(desc.SampleDesc.Count != 1)
        svDesc.ViewDimension = D3D10_SRV_DIMENSION_TEXTURE2DMS;
      else
      {
        svDesc.ViewDimension = D3D10_SRV_DIMENSION_TEXTURE2D; //MS;
        svDesc.Texture2D.MipLevels = 1;
        svDesc.Texture2D.MostDetailedMip = 0;
      }
      hr = g_device->CreateShaderResourceView( m_rt, &svDesc, m_shaderResourceView.Address());
    }
    //return hr;
  }
  
  ID3D10RenderTargetView* RenderTargetView() { return m_renderTargetView; }
  ID3D10ShaderResourceView* ShaderResourceView() { return m_shaderResourceView; }
  ID3D10Texture2D* Texture2D() { return m_rt; }
};    

CountPtr<RenderTargetTexture> g_mrts[4];
CountPtr<Texture> g_diffuse;
CountPtr<Texture> g_normalMap;

// Vertex Shader Constant Buffer ////////////////////////////////////////////////////////////////////////////////////
struct ConstantBuffer
{
  D3DXMATRIX world;
  D3DXMATRIX worldViewProj;
};


// Vertex Structure /////////////////////////////////////////////////////////////////////////////////////////////////
struct Vertex
{
  Vertex(){}
  Vertex(D3DXVECTOR3 p, D3DXVECTOR3 n, D3DXVECTOR2 t) : position(p), normal(n), texCoord0(t) { }
  
  D3DXVECTOR3 position;
  D3DXVECTOR3 normal;
  D3DXVECTOR3 tangent;
  D3DXVECTOR3 bitangent;
  D3DXVECTOR2 texCoord0;
};


// Create Depth Stencil /////////////////////////////////////////////////////////////////////////////////////////////
void CreateDepthStencil()
{
  D3D10_TEXTURE2D_DESC desc;
  desc.Width  = g_swapChainDesc.BufferDesc.Width;
  desc.Height = g_swapChainDesc.BufferDesc.Height;
  desc.MipLevels = 1;
  desc.ArraySize = 1;
  desc.Format = DXGI_FORMAT_D24_UNORM_S8_UINT;
  desc.SampleDesc.Count = g_numSamples;
  desc.SampleDesc.Quality = 0;
  desc.Usage = D3D10_USAGE_DEFAULT;
  desc.BindFlags = D3D10_BIND_DEPTH_STENCIL;
  desc.CPUAccessFlags = 0;
  desc.MiscFlags = 0;
  
  g_device->CreateTexture2D(&desc, NULL, g_depthStencilTexture.Address());
  
  D3D10_DEPTH_STENCIL_VIEW_DESC dsDesc;
  dsDesc.Format = DXGI_FORMAT_D24_UNORM_S8_UINT;
  if(desc.SampleDesc.Count != 1)
    dsDesc.ViewDimension = D3D10_DSV_DIMENSION_TEXTURE2DMS;
  else
  {
    dsDesc.ViewDimension = D3D10_DSV_DIMENSION_TEXTURE2D;
    dsDesc.Texture2D.MipSlice = 0;
  }
  
  g_device->CreateDepthStencilView( g_depthStencilTexture, &dsDesc, g_depthStencilView.Address());
}


// Create Render Target /////////////////////////////////////////////////////////////////////////////////////////////
HRESULT CreateRenderTarget(DWORD width, DWORD height, DXGI_FORMAT format, ID3D10Texture2D **rt, ID3D10RenderTargetView **view)
{
  D3D10_TEXTURE2D_DESC desc;
  desc.Width  = width;
  desc.Height = height;
  desc.MipLevels = 1;
  desc.ArraySize = 1;
  desc.Format = format;
  desc.SampleDesc.Count = g_numSamples;
  desc.SampleDesc.Quality = 0;
  desc.Usage = D3D10_USAGE_DEFAULT;
  desc.BindFlags = D3D10_BIND_RENDER_TARGET | D3D10_BIND_SHADER_RESOURCE;
  desc.CPUAccessFlags = 0;
  desc.MiscFlags = 0;
  
  
  HRESULT hr = g_device->CreateTexture2D(&desc, NULL, rt);
  
  if(SUCCEEDED(hr))
  {
    D3D10_RENDER_TARGET_VIEW_DESC rtDesc;
    rtDesc.Format = desc.Format;
    if(desc.SampleDesc.Count != 1)
      rtDesc.ViewDimension = D3D10_RTV_DIMENSION_TEXTURE2DMS;
    else
    {
      rtDesc.ViewDimension = D3D10_RTV_DIMENSION_TEXTURE2D;
      rtDesc.Texture2D.MipSlice = 0;
    }
    hr = g_device->CreateRenderTargetView( *rt, &rtDesc, view);
  }
  return hr;
}


// Create Render Targets ////////////////////////////////////////////////////////////////////////////////////////////
void CreateRenderTargets(DWORD w, DWORD h)
{
  CreateDepthStencil();
  
  g_mrts[0] = new RenderTargetTexture(w, h, DXGI_FORMAT_R16G16B16A16_FLOAT);
  g_mrts[1] = new RenderTargetTexture(w, h, DXGI_FORMAT_R10G10B10A2_UNORM);
  g_mrts[2] = new RenderTargetTexture(w, h, DXGI_FORMAT_R8G8B8A8_UNORM);
  g_mrts[3] = new RenderTargetTexture(w, h, DXGI_FORMAT_R32_FLOAT);
  
}


// Create Constant Buffer ///////////////////////////////////////////////////////////////////////////////////////////
HRESULT CreateConstantBuffer(DWORD size, ID3D10Buffer** cb)
{
  D3D10_BUFFER_DESC cbDesc;
  cbDesc.ByteWidth = size; 
  cbDesc.Usage = D3D10_USAGE_DYNAMIC;
  cbDesc.BindFlags = D3D10_BIND_CONSTANT_BUFFER;
  cbDesc.CPUAccessFlags = D3D10_CPU_ACCESS_WRITE;
  cbDesc.MiscFlags = 0;
  return g_device->CreateBuffer(&cbDesc, NULL, cb);
}


// Create Vertex Buffer /////////////////////////////////////////////////////////////////////////////////////////////
HRESULT CreateVertexBuffer(void *initialData, DWORD size, ID3D10Buffer** vb)
{
  D3D10_BUFFER_DESC bd;
  bd.Usage = D3D10_USAGE_DEFAULT;
  bd.ByteWidth = size;
  bd.BindFlags = D3D10_BIND_VERTEX_BUFFER;
  bd.CPUAccessFlags = 0;
  bd.MiscFlags = 0;
  
  D3D10_SUBRESOURCE_DATA vInit;
  vInit.pSysMem = initialData;
  
  return g_device->CreateBuffer( &bd, &vInit, vb);
}

// Create Index Buffer //////////////////////////////////////////////////////////////////////////////////////////////
HRESULT CreateIndexBuffer(void *initialData, DWORD size, ID3D10Buffer** vb)
{
  D3D10_BUFFER_DESC bd;
  bd.Usage = D3D10_USAGE_DEFAULT;
  bd.ByteWidth = size;
  bd.BindFlags = D3D10_BIND_INDEX_BUFFER;
  bd.CPUAccessFlags = 0;
  bd.MiscFlags = 0;
  
  D3D10_SUBRESOURCE_DATA vInit;
  vInit.pSysMem = initialData;
  
  return g_device->CreateBuffer( &bd, &vInit, vb);
}

// Mesh Helper Class ////////////////////////////////////////////////////////////////////////////////////////////////
class Mesh : public Counted
{
  CountPtr<ID3D10Buffer> m_vb, m_ib;
  DWORD m_numFaces;
  
    public:
  void Set()
  {
    UINT stride = sizeof(Vertex);
    UINT offset = 0;
    g_device->IASetVertexBuffers(0, 1, m_vb.ConstAddress(), &stride, &offset);
    g_device->IASetIndexBuffer(m_ib, DXGI_FORMAT_R16_UINT, 0);
    g_device->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
  }
  
  void Draw()
  {
    g_device->DrawIndexed(m_numFaces*3, 0, 0);
  }

  static Mesh *CreateSphere(float radius, DWORD numStacks, DWORD numSlices)
  {
    DWORD numVertices = (numStacks+1)*(numSlices+1);
    DWORD numFaces = 2*numSlices*numStacks;
    
    Vertex *verts = new Vertex[numVertices];
    WORD* indices = new WORD[numFaces*3];
    
    DWORD curVert = 0;
    for(DWORD stack = 0; stack <= numStacks; stack++)
    {
      D3DXVECTOR3 startVector(0,1,0);
      float zAmount = ((float)stack)/((float)numStacks);
      float zAngle = zAmount * (float)D3DX_PI;
      D3DXMATRIX zRotation;
      D3DXMatrixRotationZ(&zRotation, zAngle);
      D3DXVec3TransformCoord(&startVector, &startVector, &zRotation);
      for(DWORD slice = 0; slice <= numSlices; slice++)
      {
        float yAmount = ((float)slice) / ((float)numSlices);
        float yAngle = yAmount * 2.0f * (float)D3DX_PI;
        D3DXMATRIX rotation;
        D3DXMatrixRotationY(&rotation, yAngle);
        D3DXVECTOR3 v = startVector;
        D3DXVec3TransformCoord(&v, &v, &rotation);
        
        verts[curVert].normal = v;
        verts[curVert].position = v * radius;
        D3DXVECTOR3 y(0,1,0);
        
        float yAmountRight = ((float)(slice+1))/((float)numSlices);
        D3DXVECTOR2 texCoordRight = D3DXVECTOR2(yAmountRight*8.0f, zAmount*8.0f);
        D3DXVec3Cross(&verts[curVert].tangent, &v, &y);
        D3DXVec3Cross(&verts[curVert].bitangent, &verts[curVert].tangent, &verts[curVert].normal);
        verts[curVert].texCoord0 = D3DXVECTOR2(yAmount*8.0f, zAmount*8.0f);
        curVert++;
      }
    }
    
    DWORD curInd = 0;
    for(DWORD stack = 0; stack < numStacks; stack++)
    {
      for(DWORD slice = 0; slice < numSlices; slice++)
      {
        WORD inds[4];
        inds[0] = (WORD)((stack*(numSlices+1)) + slice);
        inds[1] = inds[0] + 1;
        inds[2] = inds[0] + (WORD)numSlices+1;
        inds[3] = inds[2] + 1;
        
        indices[curInd++] = inds[0];
        indices[curInd++] = inds[3];
        indices[curInd++] = inds[1];
        
        indices[curInd++] = inds[0];
        indices[curInd++] = inds[2];
        indices[curInd++] = inds[3];
      }
    }
    
    Mesh *m = new Mesh();
    CreateVertexBuffer(verts, numVertices*sizeof(Vertex), m->m_vb.Address());
    CreateIndexBuffer(indices, numFaces*3*sizeof(WORD), m->m_ib.Address());
    m->m_numFaces = numFaces;
    return m;
  }
};

CountPtr<Mesh> g_mesh;


// Init SwapChain And Backbuffer ////////////////////////////////////////////////////////////////////////////////////
void InitSwapChainAndBackBuffer(HWND wnd)
{
  // We need a device and a swap chain.
  ZeroMemory( &g_swapChainDesc, sizeof(g_swapChainDesc));
  g_swapChainDesc.BufferCount = 1;
  RECT r;
  GetClientRect(wnd, &r);
  g_swapChainDesc.BufferDesc.Width = r.right-r.left;
  g_swapChainDesc.BufferDesc.Height = r.bottom-r.top;
  g_swapChainDesc.BufferDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
  g_swapChainDesc.BufferDesc.RefreshRate.Numerator = 60;
  g_swapChainDesc.BufferDesc.RefreshRate.Denominator = 1;
  g_swapChainDesc.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;
  g_swapChainDesc.OutputWindow = wnd;
  g_swapChainDesc.SampleDesc.Count = 1;
  g_swapChainDesc.SampleDesc.Quality = 0;
  g_swapChainDesc.Windowed = true;
  
  DWORD flags = 0;
  #ifdef _DEBUG
    flags = D3D10_CREATE_DEVICE_DEBUG;
  #endif

  IDXGIFactory *pDXGIFactory;
  CreateDXGIFactory(__uuidof(IDXGIFactory), (void**)&pDXGIFactory);
  // Search for a PerfHUD adapter.
  UINT nAdapter = 0;
  IDXGIAdapter* adapter = NULL;
  IDXGIAdapter* selectedAdapter = NULL;
  D3D10_DRIVER_TYPE driverType = D3D10_DRIVER_TYPE_HARDWARE;
  /*while( pDXGIFactory->EnumAdapters( nAdapter, &adapter) != DXGI_ERROR_NOT_FOUND)
  {
	  if( adapter)
	  {
		  DXGI_ADAPTER_DESC adaptDesc;
		  if( SUCCEEDED( adapter->GetDesc( &adaptDesc)))
		  {
		    //MessageBox(NULL, adaptDesc.Description, "BO", MB_OK);
			  const bool isPerfHUD = wcscmp( adaptDesc.Description, L"NVIDIA PerfHUD") == 0;
			  // Select the first adapter in normal circumstances or the PerfHUD one if it exists.
			  if( nAdapter == 0 || isPerfHUD)
				  selectedAdapter = adapter;
			  if( isPerfHUD)
			  {
				  driverType = D3D10_DRIVER_TYPE_REFERENCE;
				  flags = 0;
				}
		  }
	  }
	  ++nAdapter;
  }*/
  
  D3D10CreateDeviceAndSwapChain(selectedAdapter, driverType, NULL, flags, D3D10_SDK_VERSION, &g_swapChainDesc, g_swapChain.Address(), g_device.Address());
  
  // Nab the backbuffer so we can create a render target to it
  g_swapChain->GetBuffer(0, __uuidof( ID3D10Texture2D ), (void**)g_backBuffer.Address());

  g_device->CreateRenderTargetView(g_backBuffer, NULL, g_renderTargetView.Address());
  g_device->OMSetRenderTargets(1, g_renderTargetView.ConstAddress(), NULL);
}


// Rebuild Render Targets ///////////////////////////////////////////////////////////////////////////////////////////
void RebuildRenderTargets(int w, int h)
{
  g_swapChainDesc.BufferDesc.Width = w;
  g_swapChainDesc.BufferDesc.Height = h;
  
  // Have to release the backbuffer and render target view for the swap chain
  g_backBuffer = NULL;
  g_renderTargetView = NULL;
  
  // Resize
  g_swapChain->ResizeBuffers(g_swapChainDesc.BufferCount, w, h, g_swapChainDesc.BufferDesc.Format, 0);
  
  // Regenerate the backbuffer and render target view
  g_swapChain->GetBuffer(0, __uuidof( ID3D10Texture2D ), (void**)g_backBuffer.Address());
  g_device->CreateRenderTargetView(g_backBuffer, NULL, g_renderTargetView.Address());        
  
  // Recreate our MRTs since their sizes have changed (this also recreates the depth stencil)
  CreateRenderTargets(w, h);
  
  // Set the new viewport
  g_viewport.Width = g_swapChainDesc.BufferDesc.Width;
  g_viewport.Height = g_swapChainDesc.BufferDesc.Height;
  g_viewport.MinDepth = 0.0f;
  g_viewport.MaxDepth = 1.0f;
  g_viewport.TopLeftX = 0;
  g_viewport.TopLeftY = 0;
  g_device->RSSetViewports(1, &g_viewport);
}


// Rebuild Effect ///////////////////////////////////////////////////////////////////////////////////////////////////
void RebuildEffect()
{
  D3D10_INPUT_ELEMENT_DESC layout[] =
  {
    { "SV_Position", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0,  0*sizeof(float), D3D10_INPUT_PER_VERTEX_DATA, 0 },
    { "Normal",      0, DXGI_FORMAT_R32G32B32_FLOAT, 0,  3*sizeof(float), D3D10_INPUT_PER_VERTEX_DATA, 0 },
    { "Tangent",     0, DXGI_FORMAT_R32G32B32_FLOAT, 0,  6*sizeof(float), D3D10_INPUT_PER_VERTEX_DATA, 0 },
    { "Bitangent",   0, DXGI_FORMAT_R32G32B32_FLOAT, 0,  9*sizeof(float), D3D10_INPUT_PER_VERTEX_DATA, 0 },
    { "TexCoord",    0, DXGI_FORMAT_R32G32_FLOAT,    0, 12*sizeof(float), D3D10_INPUT_PER_VERTEX_DATA, 0 }
  };
  

  DWORD flags = 0;
  #ifdef _DEBUG
    flags = D3D10_SHADER_DEBUG;
  #endif

  char numSampleString[30];
  D3D10_SHADER_MACRO defines[] = { {NULL, NULL}, {NULL,NULL} };
  if(g_numSamples > 1)
  {
    sprintf_s(numSampleString, 30, "%i", g_numSamples);
    defines[0].Name = "MULTISAMPLE";
    defines[0].Definition = numSampleString;
  }
  
  D3DX10CreateEffectFromFile(L"main.fx", defines, NULL, "fx_4_0", flags, 0, g_device, NULL, NULL, g_effect.Address(), NULL, NULL);
  
  g_geometryTechnique = g_effect->GetTechniqueByName("Geometry");
  g_testTechnique = g_effect->GetTechniqueByName("MultisampleTest");
  g_lightTechnique = g_effect->GetTechniqueByName("LightRender");

  g_effect->GetConstantBufferByName("Basic")->SetConstantBuffer(g_constantBuffer);  
  g_effect->GetConstantBufferByName("Light")->SetConstantBuffer(g_lightConstantBuffer);

  D3D10_PASS_DESC PassDesc;
  g_geometryTechnique->GetPassByIndex( 0 )->GetDesc( &PassDesc );

  g_device->CreateInputLayout(layout, countof(layout), PassDesc.pIAInputSignature, PassDesc.IAInputSignatureSize, g_layout.Address());
}


// Init D3D10 ///////////////////////////////////////////////////////////////////////////////////////////////////////  
void InitD3D10(HWND wnd)
{
  InitSwapChainAndBackBuffer(wnd);
  CreateRenderTargets(g_swapChainDesc.BufferDesc.Width, g_swapChainDesc.BufferDesc.Height);
  
  // Setup our viewport
  g_viewport.Width = g_swapChainDesc.BufferDesc.Width;
  g_viewport.Height = g_swapChainDesc.BufferDesc.Height;
  g_viewport.MinDepth = 0.0f;
  g_viewport.MaxDepth = 1.0f;
  g_viewport.TopLeftX = 0;
  g_viewport.TopLeftY = 0;
  g_device->RSSetViewports(1, &g_viewport);
  
  // Set up the constant buffers
  CreateConstantBuffer(sizeof(ConstantBuffer), g_constantBuffer.Address()); 
  CreateConstantBuffer(sizeof(LightConstantBuffer), g_lightConstantBuffer.Address());
  
  g_device->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_TRIANGLELIST);

  g_mesh = Mesh::CreateSphere(3.0f, 64, 64);
  g_mesh->Set();
  
  g_diffuse = new Texture(L"Brick1.dds");
  g_normalMap = new Texture(L"BrickBump.dds");
  ID3D10ShaderResourceView * resources[] = { g_diffuse->View(), g_normalMap->View() };
  g_device->PSSetShaderResources(0, 2, resources);
  
  RebuildEffect();
}

// WinProc //////////////////////////////////////////////////////////////////////////////////////////////////////////
LRESULT WINAPI WindowProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam )
{
  switch( message )
  {
    case WM_CREATE:
      break;

    case WM_KEYDOWN:
      switch( wParam )
      {
        case VK_F5:
        {
          if(g_numSamples == 1)
          {
            g_numSamples = 4;
          }
          else
            g_numSamples = 1;
            
          RECT r;
          GetClientRect(hWnd, &r);
          RebuildRenderTargets(r.right, r.bottom);
          RebuildEffect();
          break;
        }
        case VK_ESCAPE:
        case VK_F12:
          PostMessage(hWnd, WM_CLOSE, 0, 0);
          break;
      }
      break;
    case WM_SIZE:
    {
      if(g_device != NULL)
      {
        RECT r;
        GetClientRect(hWnd, &r);
        RebuildRenderTargets(r.right,r.bottom);
      }
      break;
    }
    case WM_DESTROY:
      PostQuitMessage( 0 );
      break;
  }
  return DefWindowProc(hWnd, message, wParam, lParam);
}


// Do Init //////////////////////////////////////////////////////////////////////////////////////////////////////////
static BOOL DoInit( HINSTANCE hInstance, int nCmdShow )
{
  HWND                hwnd;
  WNDCLASS wc;

  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = WindowProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = hInstance;
  wc.hIcon = LoadIcon( hInstance, IDI_APPLICATION );
  wc.hCursor = LoadCursor( NULL, IDC_ARROW );
  wc.hbrBackground = NULL;
  wc.lpszMenuName = NAME;
  wc.lpszClassName = NAME;
  RegisterClass( &wc );

  hwnd = CreateWindowEx(0, NAME, TITLE, WS_OVERLAPPEDWINDOW, 100, 100, 640, 480, NULL, NULL, hInstance, NULL );
  if( !hwnd )
    return FALSE;
  ShowWindow( hwnd, nCmdShow );
  UpdateWindow( hwnd );

  InitD3D10(hwnd);
  
  return TRUE;
} /* doInit */


// WinMain //////////////////////////////////////////////////////////////////////////////////////////////////////////
int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE, LPSTR, int nCmdShow)
{
  MSG msg;
  if( !DoInit( hInstance, nCmdShow ) )
    return FALSE;
  for(;;)
  {
    if(PeekMessage(&msg,NULL, 0, 0, PM_NOREMOVE))
    {
      if(!GetMessage(&msg, NULL, 0, 0))
        break;
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
    else
    {
      ConstantBuffer* cb;
      g_constantBuffer->Map( D3D10_MAP_WRITE_DISCARD, 0, (void**)&cb);
      D3DXMatrixTranslation(&cb->world, 0,0,6);
      D3DXMatrixPerspectiveFovLH(&cb->worldViewProj, (float)D3DXToRadian(60.0f), (float)g_viewport.Width/(float)g_viewport.Height, 1.0f, 1000.0f);
      D3DXMatrixMultiply(&cb->worldViewProj, &cb->world, &cb->worldViewProj);
      D3DXMatrixTranspose(&cb->world, &cb->world);
      D3DXMatrixTranspose(&cb->worldViewProj, &cb->worldViewProj);
      g_constantBuffer->Unmap();
    
      // First thing first...Clear all of those delicious render targets
      float black[] = {0,0,0,0};
      g_device->ClearRenderTargetView(g_renderTargetView, black);      
      g_device->ClearRenderTargetView(g_mrts[0]->RenderTargetView(), black);
      g_device->ClearRenderTargetView(g_mrts[1]->RenderTargetView(), black);
      g_device->ClearRenderTargetView(g_mrts[2]->RenderTargetView(), black);
      g_device->ClearDepthStencilView(g_depthStencilView, D3D10_CLEAR_DEPTH | D3D10_CLEAR_STENCIL, 1.0f, 0);
      
      // MRT Render
      g_geometryTechnique->GetPassByIndex(0)->Apply(0);
      ID3D10RenderTargetView* rts[] = {g_mrts[0]->RenderTargetView(), 
                                       g_mrts[1]->RenderTargetView(),
                                       g_mrts[2]->RenderTargetView(),
                                       g_mrts[3]->RenderTargetView()};
      g_device->IASetInputLayout(g_layout);
      
      ID3D10ShaderResourceView* views[] =
      {
        g_diffuse->View(),
        g_normalMap->View(),
        NULL
      };

      g_device->PSSetShaderResources(0, countof(views), views);
      g_device->OMSetRenderTargets(4, rts, g_depthStencilView);
      
      g_mesh->Set();
      g_mesh->Draw();
      
      
      // TestRender
      views[0] = NULL;
      views[1] = g_mrts[2]->ShaderResourceView();
      views[2] = g_mrts[3]->ShaderResourceView();

      g_device->IASetInputLayout(NULL);
      
      if(g_numSamples > 1)
      {
        rts[0] = g_mrts[1]->RenderTargetView();
        g_device->OMSetRenderTargets(1, rts, NULL);
        
        g_testTechnique->GetPassByIndex(0)->Apply(0);
        g_device->PSSetShaderResources(0, countof(views), views);
        g_device->Draw(6, 0);
      }

      
      // Lights Render
      g_device->OMSetRenderTargets(1, g_renderTargetView.ConstAddress(), NULL);
      views[0] = g_mrts[1]->ShaderResourceView();
      g_lightTechnique->GetPassByIndex(0)->Apply(0);
      g_device->PSSetShaderResources(0, countof(views), views);

      LightData lights[] =
      {
        LightData::Ambient(D3DXVECTOR3(0.2f,0.2f,0.2f)),
        LightData::Directional(D3DXVECTOR3(-1, 0, 0), D3DXVECTOR3(0.4f,0.4f, 0.4f)),
        LightData::Directional(D3DXVECTOR3( 0, 1, 0), D3DXVECTOR3(0.5f,0.5f, 0.8f)),
        LightData::Directional(D3DXVECTOR3( 1, 0, 0), D3DXVECTOR3(1,   0,    0   )),
        LightData::Directional(D3DXVECTOR3( 0,-1, 0), D3DXVECTOR3(0   ,1,    0   )),
        /*LightData::Directional(D3DXVECTOR3( 0, 1, 0), D3DXVECTOR3(0.5f,0.5f, 0.8f)),
        LightData::Directional(D3DXVECTOR3( 1, 0, 0), D3DXVECTOR3(1,   0,    0   )),
        LightData::Directional(D3DXVECTOR3( 0,-1, 0), D3DXVECTOR3(0   ,1,    0   )),
        LightData::Directional(D3DXVECTOR3(-1, 0, 0), D3DXVECTOR3(0.4f,0.4f, 0.4f)),
        LightData::Directional(D3DXVECTOR3( 0, 1, 0), D3DXVECTOR3(0.5f,0.5f, 0.8f)),
        LightData::Directional(D3DXVECTOR3( 1, 0, 0), D3DXVECTOR3(1,   0,    0   )),
        LightData::Directional(D3DXVECTOR3( 0,-1, 0), D3DXVECTOR3(0   ,1,    0   )),
        LightData::Directional(D3DXVECTOR3(-1, 0, 0), D3DXVECTOR3(0.4f,0.4f, 0.4f)),
        LightData::Directional(D3DXVECTOR3( 0, 1, 0), D3DXVECTOR3(0.5f,0.5f, 0.8f)),
        LightData::Directional(D3DXVECTOR3( 1, 0, 0), D3DXVECTOR3(1,   0,    0   )),
        LightData::Directional(D3DXVECTOR3( 0,-1, 0), D3DXVECTOR3(0   ,1,    0   )),
        LightData::Directional(D3DXVECTOR3(-1, 0, 0), D3DXVECTOR3(0.4f,0.4f, 0.4f)),
        LightData::Directional(D3DXVECTOR3( 0, 1, 0), D3DXVECTOR3(0.5f,0.5f, 0.8f)),
        LightData::Directional(D3DXVECTOR3( 1, 0, 0), D3DXVECTOR3(1,   0,    0   )),
        LightData::Directional(D3DXVECTOR3( 0,-1, 0), D3DXVECTOR3(0   ,1,    0   )),
        LightData::Directional(D3DXVECTOR3(-1, 0, 0), D3DXVECTOR3(0.4f,0.4f, 0.4f)),
        LightData::Directional(D3DXVECTOR3( 0, 1, 0), D3DXVECTOR3(0.5f,0.5f, 0.8f)),
        LightData::Directional(D3DXVECTOR3( 1, 0, 0), D3DXVECTOR3(1,   0,    0   )),
        LightData::Directional(D3DXVECTOR3( 0,-1, 0), D3DXVECTOR3(0   ,1,    0   )),
        LightData::Directional(D3DXVECTOR3(-1, 0, 0), D3DXVECTOR3(0.4f,0.4f, 0.4f)),
        LightData::Directional(D3DXVECTOR3( 0, 1, 0), D3DXVECTOR3(0.5f,0.5f, 0.8f)),*/
      };

      DWORD i = 0;
      
      // Render a pass for each batch of 8 lights
      for(i = 0; i < (countof(lights) &(~7)); i+=8)
      {
        LightConstantBuffer *lcb;
        g_lightConstantBuffer->Map( D3D10_MAP_WRITE_DISCARD, 0, (void**)&lcb);
        lcb->numLights = 8;
        lcb->lights[0] = lights[i+0];
        lcb->lights[1] = lights[i+1];
        lcb->lights[2] = lights[i+2];
        lcb->lights[3] = lights[i+3];
        lcb->lights[4] = lights[i+4];
        lcb->lights[5] = lights[i+5];
        lcb->lights[6] = lights[i+6];
        lcb->lights[7] = lights[i+7];
        g_lightConstantBuffer->Unmap();
        g_device->Draw(6, 0);
      }
      
      // Render the stragglers
      if(i < countof(lights))
      {
        LightConstantBuffer *lcb;
        g_lightConstantBuffer->Map( D3D10_MAP_WRITE_DISCARD, 0, (void**)&lcb);
        lcb->numLights = countof(lights)-i;
        for(int j = 0; j < lcb->numLights; j++)
          lcb->lights[j] = lights[i+j];
        g_lightConstantBuffer->Unmap();
        g_device->Draw(6, 0);
      }
      g_swapChain->Present( 0, 0 );
    }
  }
  
  return msg.wParam;
}
